import esbuild__default from 'esbuild';

interface BuildExtensions extends Pick<esbuild__default.PluginBuild, 'onLoad'> {
    getLoader(id: string): esbuild__default.Loader | undefined;
    isEmittedPath(id: string): boolean;
    resolveLocallyFirst(path: string, options: esbuild__default.ResolveOptions): Promise<esbuild__default.ResolveResult>;
    onTransform(options: OnTransformOptions, callback: (args: OnTransformArgs) => Promise<OnTransformResult | null | undefined>): void;
    emitChunk(options: EmitChunkOptions): Promise<Chunk>;
    emitFile(path: string, contents?: string | Buffer): Promise<File>;
    RESOLVED_AS_FILE: 1;
    RESOLVED_AS_MODULE: 2;
}
interface OnTransformOptions extends Omit<esbuild__default.OnLoadOptions, 'filter'> {
    filter?: RegExp;
    loaders?: string[];
    extensions?: string[];
}
interface OnTransformArgs extends esbuild__default.OnLoadArgs {
    code: string;
    loader: esbuild__default.Loader;
}
interface OnTransformResult {
    code?: string;
    map?: any;
    loader?: esbuild__default.Loader;
    resolveDir?: string;
    errors?: esbuild__default.Message[];
    warnings?: esbuild__default.Message[];
    watchFiles?: string[];
    pluginData?: Record<string, any>;
}
interface EmitChunkOptions extends Omit<esbuild__default.BuildOptions, 'chunkName' | 'entryNames' | 'globalName' | 'metafile' | 'outfile'> {
    path: string;
    contents?: string | Buffer;
}
interface Chunk extends esbuild__default.BuildResult {
    id: string;
    path: string;
    filePath: string;
    metafile: esbuild__default.Metafile;
}
interface File extends esbuild__default.BuildResult {
    id: string;
    path: string;
    filePath: string;
    metafile: esbuild__default.Metafile;
}

export { BuildExtensions as B, Chunk as C, EmitChunkOptions as E, File as F, OnTransformOptions as O, OnTransformArgs as a, OnTransformResult as b };
